(* A unit for read an AV file using ffmpeg APIs.  Based of tutorial C program written by Martin Boehme and Pascal Header files ported by Victor Zinetz to Windows Object Pascal.

This source is a LGPL product by The Creative CAT *)

Unit AVRead;

interface
uses avutil,      (* av_free *)
     mathematics, (* av_rescale_q *) 
     avcodec, avformat, rational;

type
  VideoFrameDecoder = object
    public
      filename : string;             (* Pascal str + null terminated *)
      videoStream : longint;
      pFrame : PAVFrame; 	     (* pointer to decoded frame *)
      pDestFrame : PAVFrame; 	     (* pointer to destination frame *)
      width, height : longint; 	     (* Frame size *)
      availFrame,	 	     (* decoded successfully? *)
      success : boolean;	     (* done successfully? *)
      constructor init(fname : string; destFmt : TAVPixelFormat; strm : longint);
      procedure GetNextFrame;        (* Get new frame from the file *)
      procedure ConvertFrame;        (* Convert pFrame to pDestFrame *)
      procedure Seek(seek_target : int64);      (* seek the file *)
      function isKeyFrame : boolean; (* current frame is a keyframe *)
      function FramePTS : int64;     (* Presentation time stamp *)
      function FrameCodedNum : longint;         (* Coding/Decording time stamp *)
      function FrameDispNum : longint;
      function FrameNum : longint;   (* Frame # calculated locally *)
      function start_time : int64;   (* start time in AV_TIME_BASE unit *)
      function StartTime : double;   (* start time in sec *)
      function time_duration : int64;(* time duration in AV_TIME_BASE unit *)
      function DurationInSec : double;  (* time duration in sec *)
      procedure get_frame_rate(var frate : TAVRational); (* frame rate *)
      function FrameRate : double;   (* fps *)
      function TotalFrames : int64;  (* Number of frames *)
      function Language : string;    (* Language *)
      procedure time_base(var tbase : TAVRational); (* time base = 1/fps *)
      function TimeBase : double;    (* 1/fps *)
      procedure dumpStreamParam;
      procedure videoStreamParamLocalDump(var t : system.text);
      procedure FormatCtxLocalDump(var t : system.text);
      destructor done;
    private
      pFormatCtx : PAVFormatContext;    (* pointer to format context *)
      pCodecCtx : PAVCodecContext;      (* pointer to codec context *)
      pCodec : PAVCodec;                (* pointer to codec *)
      destFrameFormat : TAVPixelFormat; (* format of destination frame *)
      packet : TAVPacket;
      bytesRemaining : longint;
      buf : array of byte;
      maxDTSread,
      fnum : longint;					(* number of frame read *)
      procedure StreamParamLocalDump(var t : system.text; stream : integer);
      procedure BriefStreamParamDump(var t : system.text; stream : integer);
  end;
  PVideoFrameDecoder = ^VideoFrameDecoder;

var
  decoder : VideoFrameDecoder;

implementation

procedure FatalError(s : string);
begin
  writeln(stderr, s);
  halt(-1)
end;

constructor VideoFrameDecoder.init(fname : string; destFmt : TAVPixelFormat; 
								   strm : longint);
var numBytes : longint; (* working buffer size *)
    i : integer;
begin
  filename := fname + #0;    // force it be a zero-terminate string
  destFrameFormat := destFmt;

(* Register all formats and codecs. Calling it once, it is enough. *)
  av_register_all();

(* Open an AV file *)
  (* nil, 0, nil - using values automatically set by the library *)
  if av_open_input_file(pFormatCtx, @filename[1], nil, 0, nil) <> 0 then
    FatalError('cannot open ' + filename);

(*  Retrieve stream information *)
  if av_find_stream_info(pFormatCtx) < 0 then
    FatalError('cannot find stream information');

(* Dump AV info to standard error *)
  dump_format(pFormatCtx, 0, @filename[1], ord(false));

(* Find the first video stream, currently finds the LAST one for it cannot
   handle RX3-1.mpg, RX3-2.mpg *)
  packet.data := nil;      (* init global var *)  
  bytesRemaining := 0;     (* init global var *) 

  if strm = -1 then begin
    videoStream := -1;       (* find stream containing video *)
    for i := 0 to pred(pFormatCtx^.nb_streams) do begin
      if pFormatCtx^.streams[i]^.codec^.codec_type = CODEC_TYPE_VIDEO then begin
        videoStream := i;
        break
      end
    end
  end else
    videoStream := strm;

  if videoStream = -1 then
    FatalError('video stream cannot be found')
  else
    writeln('decode stream #', videoStream ,' with codec ',
			pFormatCtx^.streams[videoStream]^.codec^.codec_name);

(* Get a pointer to the codec context for the video stream *)
  pCodecCtx := pFormatCtx^.streams[videoStream]^.codec;

(* Find the decoder for the video stream *)
  pCodec := avcodec_find_decoder(pCodecCtx^.codec_id);
  if pCodec = nil then FatalError('codec not found');

(* Open codec *)
  if avcodec_open(pCodecCtx, pCodec) < 0 then FatalError('codec cannot be opened');

(* Allocate a buffer to store a video frame *)
  pFrame := avcodec_alloc_frame;

  if destFrameFormat <> PIX_FMT_NONE then begin
    (* Allocate another frame as an AVFrame structure to store coverted picture *)
    pDestFrame := avcodec_alloc_frame;
    if pDestFrame = nil then FatalError('cannot allocate an AVFrame');

    (* Set up destination *)
    with pCodecCtx^ do begin
      (* Allocate working buffer - get required buffer size and allocate it *)
      numBytes := avpicture_get_size(destFrameFormat, width, height);
      SetLength(buf, numBytes);  // using FPC's dynamic array

      (* Assign appropriate parts of buffer to image planes in pFrameRGB *)
      avpicture_fill(PAVPicture(pDestFrame), @buf[0], destFrameFormat, width, height)
    end
  end else begin
    pDestFrame := nil;
    SetLength(buf, 0)
  end;

(* Copy frame size to public vars *)
  width  := pCodecCtx^.width;
  height := pCodecCtx^.height;

(* initialise frame counter *)
  fnum := 0;
  maxDTSread := -1;
  availFRame := false
end;

procedure VideoFrameDecoder.GetNextFrame;
(* Get a decoded frame from packets.  Returns availability *)
var frameFinished : longint = 0;
    dts : integer;
begin
  while av_read_frame(pFormatCtx, packet) >= 0 do begin
    if packet.stream_index = videoStream then begin // target video stream found
      avcodec_decode_video(pCodecCtx, pFrame, frameFinished, 
			   packet.data, packet.size);  			// Decode a video frame
      if frameFinished <> 0 then begin  			// we get a video frame now
        av_free_packet(packet);						// free the last packet
	break
      end
    end;
    av_free_packet(packet)							// free the packet to read again
  end;
  availFrame := (frameFinished <> 0);
  if fnum >= 0 then (* framed have been read continuously, so fnum keeps its consistency *)
    inc(fnum, ord(availFrame))
  else                                        (* fnum is not available *)
    if availFrame then begin
      dts := FrameCodedNum;
      if dts > maxDTSread then maxDTSread := dts;  (* catch up *)
      if isKeyframe then fnum := maxDTSread   (* update fnum *)
    end
end;

procedure VideoFrameDecoder.ConvertFrame;
begin
  if pDestFrame <> nil then
    with pCodecCtx^ do
      img_convert(PAVPicture(pDestFrame), destFrameFormat, PAVPicture(pFrame), 
                  pix_fmt, width, height)
end;

procedure VideoFrameDecoder.Seek(seek_target : int64);
begin
  success := false;
  if videoStream >= 0 then begin
    seek_target := av_rescale_q(seek_target, AV_TIME_BASE_Q,
                      pFormatCtx^.streams[videoStream]^.time_base);
	(* av_rescale_q(a, b, c : int64) calculates a * b / c *)
    success := av_seek_frame(pFormatCtx, videoStream, seek_target, 0) >= 0;
    if not success then writeln(stderr, 'error while seeking')
  end
end;

procedure VideoFrameDecoder.dumpStreamParam;
type
  a = array[1..250] of byte;
var
  t : text;
  p : ^a;
  i : integer;
begin
  p := pointer(pFormatCtx^.streams[videoStream]);
  assign(t, 't.txt');
  rewrite(t);
  for i:= 1 to sizeof(TAVStream) do begin
    write(t, i:4,  p^[i] :4, ' ');
    if p^[i] in [ord(' ')..ord('z')] then writeln(t, chr(p^[i]))
                                     else writeln(t)
  end;
  close(t)
end;

function  VideoFrameDecoder.isKeyFrame : boolean;
begin
  isKeyFrame := pFrame^.key_frame = 1
end;

function  VideoFrameDecoder.FrameNum : longint;
begin
  FrameNum := fnum
end;

function  VideoFrameDecoder.FramePTS : int64;
begin
  FramePTS := pFrame^.pts
end;

function  VideoFrameDecoder.FrameCodedNum : longint;
begin
  FrameCodedNum := pFrame^.coded_picture_number
end;

function  VideoFrameDecoder.FrameDispNum : longint;
begin
  FrameDispNum := pFrame^.display_picture_number
end;

function  VideoFrameDecoder.start_time : int64;
begin
  start_time := pFormatCtx^.streams[videoStream]^.start_time
end;

function  VideoFrameDecoder.StartTime : double;
begin
  StartTime := pFormatCtx^.streams[videoStream]^.start_time / AV_TIME_BASE
end;

function  VideoFrameDecoder.time_duration : int64;
begin
  time_duration := pFormatCtx^.streams[videoStream]^.duration
end;

function  VideoFrameDecoder.DurationInSec : double;
begin
  DurationInSec := pFormatCtx^.streams[videoStream]^.duration / AV_TIME_BASE
end;

procedure VideoFrameDecoder.get_frame_rate(var frate : TAVRational);
begin
  frate := pFormatCtx^.streams[videoStream]^.r_frame_rate
end;

function  VideoFrameDecoder.FrameRate : double;
begin
  with pFormatCtx^.streams[videoStream]^.r_frame_rate do
    FrameRate := num / den
end;

function  VideoFrameDecoder.TotalFrames : int64;
begin
  TotalFrames := pFormatCtx^.streams[videoStream]^.nb_frames
end;

function  VideoFrameDecoder.Language : string;
var s: string[3];
begin
  s[0] := chr(3);
  with pFormatCtx^.streams[videoStream]^ do begin
    s[1] := language[0];
    s[2] := language[1];
    s[3] := language[2]
  end;
  Language := s
end;

procedure VideoFrameDecoder.time_base(var tbase : TAVRational);
begin
  tbase := pFormatCtx^.streams[videoStream]^.time_base
end;

function  VideoFrameDecoder.TimeBase : double;
var s1, s2 : string;
begin
  with pFormatCtx^.streams[videoStream]^.time_base do
    TimeBase := num / den
end;

procedure VideoFrameDecoder.BriefStreamParamDump(var t : system.text; stream : integer);
begin
  with pFormatCtx^.streams[stream]^ do begin
    writeln(t, 'index:', index, ' id:', id);
    if codec <> nil then
      with codec^ do begin
        write  (t, 'Codec Context codec_name: ', codec_name,
                   ' codec_type: ', ord(codec_type),
                   ' codec_id: ', ord(codec_id),
                   ' time_base: ', time_base.num, '/', time_base.den);
        if stream = videoStream then
          write(t, ' picture size: ', width, 'x', height,
                   ' gop_size :', gop_size,
                   ' pix_fmt: ', ord(pix_fmt));
        writeln(t)
      end
  end
end;

procedure VideoFrameDecoder.StreamParamLocalDump(var t : system.text; stream : integer);
begin
  writeln(t, 'Parametres of stream #', stream);
  with pFormatCtx^.streams[stream]^ do begin
    BriefStreamParamDump(t, stream);
    writeln(t, 'real base frame rate: ', r_frame_rate.num, '/', r_frame_rate.den);
    (* priv_data: pointer; *)
    writeln(t, 'codec_info_duration: ', codec_info_duration,
              ' codec_info_nb_frames: ', codec_info_nb_frames);
    writeln(t, 'pts: ', pts.val, '/', pts.num, '/', pts.den);
    writeln(t, 'time_base: ', time_base.num, '/', time_base.den);
    writeln(t, 'pts_wrap_bits: ', pts_wrap_bits,
              ' stream_copy :', stream_copy,
              ' discard: ', ord(discard));
    writeln(t, 'quality: ', quality);
    writeln(t, 'start_time: ', start_time, ' duration: ', duration);
    
    writeln(t, 'language: ', language);
    writeln(t, 'need_parsing: ', ord(need_parsing));

    writeln(t, 'cur_dts: ', cur_dts,
              ' last_IP_duration: ', last_IP_duration,
              ' last_IP_pts: ', last_IP_pts);
    (* index_entries: PAVIndexEntry; *)
    writeln(t, 'nb_index_entries: ', nb_index_entries,
              ' index_entries_allocated_size: ', index_entries_allocated_size);
    writeln(t, 'nb_frames: ', nb_frames);
    (* pts_buffer: array [0..MAX_REORDER_DELAY] of int64 *)
  end
end;

procedure VideoFrameDecoder.videoStreamParamLocalDump(var t : system.text);
begin
  StreamParamLocalDump(t, videoStream)
end;

procedure VideoFrameDecoder.FormatCtxLocalDump(var t : system.text);
var i : integer;
begin
  writeln(t, 'Parametres of Format Context');
  with pFormatCtx^ do begin
    if av_class <> nil then
      with av_class^ do
        writeln(t, 'AVClass class_name: ', class_name);
  (*  iformat: PAVInputFormat;
    oformat: PAVOutputFormat;
    priv_data: pointer;
    pb: TByteIOContext; *)
    writeln(t, 'nb_streams: ', nb_streams);
    for i := 1 to nb_streams do begin
      writeln('Stream #', pred(i));
      BriefStreamParamDump(t, pred(i))
    end;
    writeln(t, 'filename: ', filename, ' timestamp: ', timestamp);
    writeln(t, 'title: ', title);
    writeln(t, 'author: ', author);
    writeln(t, 'copyright: ', copyright);
    writeln(t, 'comment: ', comment);
    writeln(t, 'album: ', album);
    writeln(t, 'year: ', year,
              ' track: ', track,
              ' genre: ', genre);

    writeln(t, 'ctx_flags: ', ctx_flags);
(*    packet_buffer: PAVPacketList;*)

    writeln(t, 'start_time: ', start_time,
              ' duration: ', duration,
              ' file_size: ', file_size,
              ' bit_rate: ', bit_rate)
{
    (* av_read_frame() support *)
    cur_st: PAVStream;
    cur_ptr: pbyte;
    cur_len: integer;
    cur_pkt: TAVPacket;

    (* av_seek_frame() support *)
    data_offset: int64; (* offset of the first packet *)
    index_built: integer;

    mux_rate: integer;
    packet_size: integer;
    preload: integer;
    max_delay: integer;

    (* number of times to loop output in formats that support it *)
    loop_output: integer;

    flags: integer;
    loop_input: integer;
    (* decoding: size of data to probe; encoding unused *)
    probesize: cardinal;

    (**
     * maximum duration in AV_TIME_BASE units over which the input should be analyzed in av_find_stream_info()
     *)
    max_analyze_duration: integer;

    key: pbyte;
    keylen : integer
  end;
}
  end
end;


destructor VideoFrameDecoder.done;
begin
(* deallocate buffers *)
  SetLength(buf, 0);
  av_free(pDestFrame);   // Free the destination image
  av_free(pFrame);       // Free the YUV frame

(* close the codec *)
  avcodec_close(pCodecCtx);

(* Close the video file *)
  av_close_input_file(pFormatCtx)
end;

end.
